-module(icfp_print).

-include("icfp.hrl").

-define(MAX_RANK_LOOKAHEAD, 20).	% pretty arbitrary

-import(lists, [map/2, foldl/3, reverse/1]).
-export([file/3]).

%% Stringify the parse tree (Parse) from initial state Istate and
%% write the result to Outfile.
file(Outfile, Parse, Istate) ->
    file:write_file(Outfile, print(Parse, Istate)).

print(Parse, State) ->
    print(Parse, State, []).

%% Stack is a list of tags, like 'b', 5, 'B', 'TT', etc
print([], S, Stack) ->
    %% end of input; unwind the stack to be finished
    [format_pop(Tag) || Tag <- Stack];
print([{SNew, Text}|T], SOld, Stack0) ->
    Ops = ops(SOld, SNew),
    {Pushes, Pops, Stack} = process_ops(Ops, Stack0, SNew, T),
    PopText = map(fun(Tag) -> format_pop(Tag) end, Pops),
    PushText = map(fun(Tag) -> format_push(Tag) end, Pushes),
    [PopText, PushText, Text | print(T, SNew, reverse(Pushes) ++ Stack)].

%% Returns: {Pushes, Pops, Stack}
process_ops(Ops, SOld, SNew, Parse) ->
    Pops0 = [Tag || {pop, Tag} <- Ops],
    {Pops1, Repushes, Stack1} = process_pops(Pops0, SOld),
    Pushes0 = [Tag || {push, Tag} <- Ops],
    Pushes1 = Repushes ++ Pushes0,
    Pushes2 = rank_sort_pushes(Pushes1, Parse, SNew),
    {Pushes2, Pops1, Stack1}.

%% process_pops(PopTags, Stack) => {SortedPops, Repushes, Stack'}
%% `Repushes' are elements that were lifted to remove others beneath
%% them - they need to be pushed back on.
process_pops(Pops, Stack) ->
    process_pops(Pops, Stack, [], []).

process_pops([], Stack, Acc, Repushes) ->
    {reverse(Repushes) ++ reverse(Acc), Repushes, Stack};
process_pops(Pops, [H|T], Acc, Repushes) ->
    case lists:member(H, Pops) of
	true  -> process_pops(Pops -- [H], T, [H|Acc], Repushes);
	false -> process_pops(Pops, T, Acc, [H|Repushes])
    end.

%% Returns: Pushes'
%% Try to order Pushes so that they'll be popped off in the right order.
rank_sort_pushes(Pushes, Parse, S0) ->
    Ranked = rank_pushes(Pushes, Parse, S0, 1),
    Sorted = reverse(lists:keysort(1, Ranked)),
    map(fun({Rank, Push}) -> Push end,
	Sorted).

rank_pushes(Pushes, Parse, _, Rank) when Parse == [];
					 Rank == ?MAX_RANK_LOOKAHEAD ->
    map(fun(Push) -> {?MAX_RANK_LOOKAHEAD, Push} end,
	Pushes);
rank_pushes(Pushes, [{S,_}|T], S0, Rank) ->
    F = fun(Push, {Stops, Gos}) ->
		case stops_in(Push, S0, S) of
		    true ->
			{[Push|Stops], Gos};
		    false ->
			{Stops, [Push|Gos]}
		end
	end,
    {Stops, Gos} = foldl(F, {[],[]}, Pushes),
    map(fun(Stop) -> {Rank,Stop} end, Stops) ++ rank_pushes(Gos,T,S0,Rank+1);
rank_pushes([], _, _, _) ->
    [].

%% stops_in(Tag, TagState, NewState) => true | false
%% See if the Tag stops in NewState (i.e. must be popped to get there)
stops_in('B', _, #s{b=true})  -> false;
stops_in('B', _, #s{b=false}) -> true;
stops_in('EM', _, #s{em=false, s=false}) -> true;
stops_in('EM', _, _)                     -> false;
stops_in('I', _, #s{i=true})  -> false;
stops_in('I', _, #s{i=false}) -> true;
stops_in('S', _, #s{s=true})  -> false;
stops_in('S', _, #s{s=false}) -> true;
stops_in('TT', _, #s{tt=true})  -> false;
stops_in('TT', _, #s{tt=false}) -> true;
stops_in('U', #s{u=Old}, #s{u=New}) -> New < Old;
stops_in(N, _, #s{size=N}) when integer(N) -> false;
stops_in(N, _, _) when integer(N)          -> true;
stops_in(Color, _, #s{color=Color}) -> false;
stops_in(Color, _, _)                -> true.

%% ops(OldState, NewState) => [Op]
%% Decides which push and pop operations are needed to go from state
%% Old to state New. Returns them in a list of {push, Tag} and {pop,
%% Tag} tuples.
ops(Old, New) ->
    lists:flatmap(fun(Item) -> tag(Item, Old, New) end,
		  record_info(fields, s)).

%% tag(Field, OldState, NewState) => [Op]
tag(b, #s{b=X}, #s{b=Y}) ->
    change_bool('B', X, Y);
tag(em, #s{em=X}, #s{em=Y}) ->
    change_bool('EM', X, Y);
tag(i, #s{i=X}, #s{i=Y}) ->
    change_bool('I', X, Y);
tag(s, #s{s=X}, #s{s=Y}) ->
    change_bool('S', X, Y);
tag(tt, #s{tt=X}, #s{tt=Y}) ->
    change_bool('TT', X, Y);
tag(u, #s{u=X}, #s{u=Y}) ->
    if
	X == Y -> [];
	X > Y  -> lists:duplicate(X - Y, {pop, 'U'}); % want less underline
	X < Y  -> lists:duplicate(Y - X, {push, 'U'}) % want more underline
    end;
tag(size, #s{size=X}, #s{size=Y}) ->
    if
	X == Y         -> [];
	X == undefined -> [{push, Y}];
	Y == undefined -> [{pop, X}];
	X /= Y         -> [{pop, X}, {push, Y}]
    end;
tag(color, #s{color=X}, #s{color=Y}) ->
    if
	X == Y         -> [];
	X == undefined -> [{push, Y}];
	Y == undefined -> [{pop, X}];
	true           -> [{pop, X}, {push, Y}]
    end.

%% change_bool(Tag, Old, New)
%% Determine the Ops for setting a boolean value of Tag from Old to New
change_bool(Tag, X, X)        -> [];
change_bool(Tag, true, false) -> [{pop, Tag}];
change_bool(Tag, false, true) -> [{push, Tag}].

%% Text formatting
format_pop(Tag) ->
    ["</", tag_to_string(Tag), ">"].

format_push(Tag) ->
    ["<", tag_to_string(Tag), ">"].

tag_to_string(I) when integer(I) ->
    integer_to_list(I);
tag_to_string(A) when atom(A) ->
    atom_to_list(A).

